from abc import ABCMeta, abstractmethod
import uuid
from six import with_metaclass

from eventsourcing.domain.model.entity import EventSourcedEntity
from eventsourcing.domain.model.events import publish


class Example(EventSourcedEntity):

    class Created(EventSourcedEntity.Created):
        @property
        def a(self):
            return self.__dict__['a']

        @property
        def b(self):
            return self.__dict__['b']

    def __init__(self, a, b, **kwargs):
        super(Example, self).__init__(**kwargs)
        self._a = a
        self._b = b

    @property
    def a(self):
        return self._a

    @a.setter
    def a(self, value):
        self._set_event_sourced_attribute_value(name='_a', value=value)

    @property
    def b(self):
        return self._b

    @b.setter
    def b(self, value):
        self._set_event_sourced_attribute_value(name='_b', value=value)


class Repository(with_metaclass(ABCMeta)):
    
    @abstractmethod
    def __getitem__(self, item):
        """Returns example entity for given ID.
        """


def register_new_example(a, b):
    """
    Factory method for example entities.
    """
    entity_id = uuid.uuid4().hex
    event = Example.Created(entity_id=entity_id, a=a, b=b)
    entity = Example.mutator(self=Example, event=event)
    publish(event=event)
    return entity



import unittest
from eventsourcing.infrastructure.stored_events import InMemoryStoredEventRepository
from eventsourcing.infrastructure.event_store import EventStore
from eventsourcing.domain.model.example import Example


class TestEventStore(unittest.TestCase):

    def test(self):
        repo = InMemoryStoredEventRepository()
        event_store = EventStore(stored_event_repo=repo)

        # Check there are zero stored events in the repo.
        entity_events = repo.get_entity_events('entity1')
        self.assertEqual(0, len(entity_events))

        # Store a domain event.
        event1 = Example.Created(entity_id='entity1', a=1, b=2)
        event_store.append(event1)

        # Check there is one stored event in the repo.
        entity_events = repo.get_entity_events('entity1')
        self.assertEqual(1, len(entity_events))

        # Store another domain event.
        event1 = Example.Created(entity_id='entity1', a=1, b=2)
        event_store.append(event1)

        # Check there are two stored events in the repo.
        entity_events = repo.get_entity_events('entity1')
        self.assertEqual(2, len(entity_events))



from six import with_metaclass
from eventsourcing.domain.model.events import DomainEvent, publish, ABCMeta


class EventSourcedEntity(with_metaclass(ABCMeta)):


    class Created(DomainEvent):

        def __init__(self, entity_version=0, **kwargs):
            super(EventSourcedEntity.Created, self).__init__(entity_version=entity_version, **kwargs)

    class AttributeChanged(DomainEvent):
        pass

    class Discarded(DomainEvent):
        pass

    def __init__(self, entity_id):
        self._id = entity_id
        self._version = 0
        self._is_discarded = False

    def _increment_version(self):
        self._version += 1

    def _assert_not_discarded(self):
        assert not self._is_discarded

    @property
    def id(self):
        return self._id

    def _validate_originator(self, event):
        assert self.id == event.entity_id, (self.id, event.entity_id)
        assert self._version == event.entity_version, "{} != {}".format(self._version, event.entity_version)

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def discard(self):
        self._assert_not_discarded()
        event = self.Discarded(entity_id=self._id, entity_version=self._version)
        self._apply(event)
        publish(event)

    def _apply(self, event):
        assert isinstance(self, EventSourcedEntity)
        self.mutator(self, event)

    @staticmethod
    def mutator(self, event):
        event_type = type(event)
        if event_type == self.Created:
            assert issubclass(self, EventSourcedEntity), self
            self = self(a=event.a, b=event.b, entity_id=event.entity_id)
            self._increment_version()
            return self
        elif event_type == self.AttributeChanged:
            self._validate_originator(event)
            setattr(self, event.name, event.value)
            self._increment_version()
            return self
        elif event_type == self.Discarded:
            self._validate_originator(event)
            self._is_discarded = True
            self._increment_version()
            return None
        else:
            raise NotImplementedError(repr(event_type))

    def _set_event_sourced_attribute_value(self, name, value):
        self._assert_not_discarded()
        event = self.AttributeChanged(name=name, value=value, entity_id=self._id, entity_version=self._version)
        self._apply(event)
        publish(event)



import unittest
try:
    from unittest import mock
except:
    import mock
from eventsourcing.domain.model.events import publish, DomainEvent
from eventsourcing.infrastructure.event_store import EventStore
from eventsourcing.infrastructure.persistence_subscriber import PersistenceSubscriber


class TestPersistenceSubscriber(unittest.TestCase):

    def setUp(self):
        self.mock_event_store = mock.Mock(spec=EventStore)
        self.persistence_subscriber = PersistenceSubscriber(event_store=self.mock_event_store)

    def tearDown(self):
        self.persistence_subscriber.close()

    def test(self):
        # Check the publishing a domain event causes 'append' to be called on the event store.
        self.assertEqual(0, self.mock_event_store.append.call_count)
        publish(mock.Mock(spec=DomainEvent))
        self.assertEqual(1, self.mock_event_store.append.call_count)



import unittest
from sqlalchemy.orm.scoping import ScopedSession
from eventsourcing.application.example import ExampleApplication
from eventsourcing.application.main import EventSourcedApplication
from eventsourcing.infrastructure.event_sourced_repos.example_repo import ExampleRepository
from eventsourcing.infrastructure.event_store import EventStore
from eventsourcing.infrastructure.persistence_subscriber import PersistenceSubscriber
from eventsourcing.infrastructure.stored_events import StoredEventRepository


class TestEventSourcedApplication(unittest.TestCase):

    def test(self):
        # Setup an event sourced application, use it as a context manager.
        with EventSourcedApplication() as app:

            # Check there's a DB session.
            self.assertIsInstance(app.db_session, ScopedSession)

            # Check there's a stored event repo.
            self.assertIsInstance(app.stored_event_repo, StoredEventRepository)
            self.assertEqual(app.stored_event_repo.db_session, app.db_session)

            # Check there's an event store.
            self.assertIsInstance(app.event_store, EventStore)
            self.assertEqual(app.event_store.stored_event_repo, app.stored_event_repo)

            # Check there's a persistence subscriber.
            self.assertIsInstance(app.persistence_subscriber, PersistenceSubscriber)
            self.assertEqual(app.persistence_subscriber.event_store, app.event_store)


class TestExampleApplication(unittest.TestCase):

    def test(self):
        # Setup the example application, use it as a context manager.
        with ExampleApplication() as app:

            # Check there's an example repository.
            self.assertIsInstance(app.example_repo, ExampleRepository)

            assert isinstance(app, ExampleApplication)  # For PyCharm...

            # Register a new example.
            example1 = app.register_new_example(a=10, b=20)

            # Check the example is available in the repo.
            entity1 = app.example_repo[example1.id]
            self.assertEqual(10, entity1.a)
            self.assertEqual(20, entity1.b)
            self.assertEqual(example1, entity1)

            # Change attribute values.
            entity1.a = 100

            # Check the new value is available in the repo.
            entity1 = app.example_repo[example1.id]
            self.assertEqual(100, entity1.a)



from eventsourcing.infrastructure.event_store import EventStore
from eventsourcing.infrastructure.persistence_subscriber import PersistenceSubscriber
from eventsourcing.infrastructure.stored_events_sqlalchemy import SQLAlchemyStoredEventRepository, \
    get_scoped_session_facade


class EventSourcedApplication(object):

    def __init__(self, db_uri=None):
        self.db_session = self.create_db_session(db_uri)
        self.stored_event_repo = self.create_stored_event_repo()
        self.event_store = self.create_event_store()
        self.persistence_subscriber = self.create_persistence_subscriber()

    @staticmethod
    def create_db_session(uri):
        return get_scoped_session_facade(uri=uri)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        self.persistence_subscriber.close()

    def create_event_store(self):
        return EventStore(stored_event_repo=self.stored_event_repo)

    def create_stored_event_repo(self):
        return SQLAlchemyStoredEventRepository(db_session=self.db_session)

    def create_persistence_subscriber(self):
        return PersistenceSubscriber(event_store=self.event_store)



from abc import ABCMeta, abstractproperty
from six import with_metaclass
from eventsourcing.infrastructure.event_player import EventPlayer


class EventSourcedRepository(with_metaclass(ABCMeta)):

    def __init__(self, event_store):
        self.event_player = self.construct_event_player(event_store)

    def construct_event_player(self, event_store):
        return EventPlayer(event_store=event_store, domain_class=self.domain_class)

    @abstractproperty
    def domain_class(self):
        """Defines the type of entity available in this repo.
        """

    def __getitem__(self, item):
        return self.event_player[item]

    def __contains__(self, item):
        try:
            self[item]
        except KeyError:
            return False
        else:
            return True



from functools import reduce
from eventsourcing.infrastructure.event_store import EventStore
from eventsourcing.infrastructure.stored_events import recreate_domain_event


class EventPlayer(object):

    def __init__(self, event_store, domain_class):
        assert isinstance(event_store, EventStore)
        self.event_store = event_store
        self.domain_class = domain_class

    def __getitem__(self, item):
        stored_events = self.event_store.stored_event_repo.get_entity_events(item)
        entity = reduce(self.domain_class.mutator, map(recreate_domain_event, stored_events), self.domain_class)
        if entity is None:
            raise KeyError
        elif entity is self.domain_class:
            raise KeyError
        else:
            return entity



from eventsourcing.domain.model.events import subscribe, DomainEvent, unsubscribe
from eventsourcing.infrastructure.event_store import EventStore


class PersistenceSubscriber(object):

    def __init__(self, event_store):
        assert isinstance(event_store, EventStore)
        self.event_store = event_store
        subscribe(self.is_domain_event, self.store_domain_event)

    @staticmethod
    def is_domain_event(event):
        return isinstance(event, DomainEvent)

    def store_domain_event(self, event):
        self.event_store.append(event)

    def close(self):
        unsubscribe(self.is_domain_event, self.store_domain_event)



from eventsourcing.domain.model.example import Repository, Example
from eventsourcing.infrastructure.event_sourced_repo import EventSourcedRepository


class ExampleRepository(EventSourcedRepository, Repository):
    """
    Event sourced repository for the Example domain model entity.
    """
    domain_class = Example



import unittest

from eventsourcing.domain.model.example import register_new_example, Example
from eventsourcing.infrastructure.event_sourced_repos.example_repo import ExampleRepository
from eventsourcing.infrastructure.event_store import EventStore
from eventsourcing.infrastructure.persistence_subscriber import PersistenceSubscriber
from eventsourcing.infrastructure.stored_events import InMemoryStoredEventRepository


class TestDomainEntity(unittest.TestCase):

    def setUp(self):
        # Setup the persistence subscriber.
        self.event_store = EventStore(InMemoryStoredEventRepository())
        self.persistence_subscriber = PersistenceSubscriber(event_store=self.event_store)

    def tearDown(self):
        self.persistence_subscriber.close()

    def test_entity_lifecycle(self):
        # Check the factory creates an instance.
        example1 = register_new_example(a=1, b=2)
        self.assertIsInstance(example1, Example)
        self.assertEqual(1, example1.a)
        self.assertEqual(2, example1.b)

        # Check a second instance with the same values is not "equal" to the first.
        example2 = register_new_example(a=1, b=2)
        self.assertNotEqual(example1, example2)

        # Setup the repo.
        repo = ExampleRepository(self.event_store)

        # Check the example entities can be retrieved from the example repository.
        entity1 = repo[example1.id]
        self.assertIsInstance(entity1, Example)
        self.assertEqual(1, entity1.a)
        self.assertEqual(2, entity1.b)

        entity2 = repo[example2.id]
        self.assertIsInstance(entity2, Example)
        self.assertEqual(1, entity2.a)
        self.assertEqual(2, entity2.b)

        # Check the entity can be updated.
        entity1.a = 100
        self.assertEqual(100, repo[entity1.id].a)
        entity1.b = -200
        self.assertEqual(-200, repo[entity1.id].b)

        # Check the entity can be discarded.
        entity1.discard()
        self.assertRaises(KeyError, repo.__getitem__, entity1.id)

    def test_mutator_not_implemented_error(self):
        self.assertRaises(NotImplementedError, Example.mutator, Example, None)


from abc import ABCMeta
import itertools
from six import with_metaclass
from eventsourcing.utils.time import utc_now


if not hasattr(object, '__qualname__'):
    class QualnameABCMeta(ABCMeta):
        """Supplies __qualname__ to object classes with this metaclass.
        """
        __outer_classes = {}

        def __init__(cls, name, bases, dict):
            super(QualnameABCMeta, cls).__init__(name, bases, dict)
            QualnameABCMeta.__outer_classes[cls] = None
            for class_attr in dict.values():  # iterate class attributes
                for outer_class in QualnameABCMeta.__outer_classes:
                    if class_attr == outer_class:  # is the object an already registered type?
                        QualnameABCMeta.__outer_classes[outer_class] = cls
                        break

        @property
        def __qualname__(cls):
            c = QualnameABCMeta.__outer_classes[cls]
            if c is None:
                return cls.__name__
            else:
                return "%s.%s" % (c.__qualname__, cls.__name__)

    ABCMeta = QualnameABCMeta


class DomainEvent(with_metaclass(ABCMeta)):

    def __init__(self, entity_id=None, entity_version=None, timestamp=None, **kwargs):
        assert entity_id is not None
        assert entity_version is not None
        self.__dict__['entity_id'] = entity_id
        self.__dict__['entity_version'] = entity_version
        self.__dict__['timestamp'] = utc_now() if timestamp is None else timestamp
        self.__dict__.update(kwargs)

    def __setattr__(self, key, value):
        raise AttributeError("DomainEvent attributes are read-only")

    @property
    def entity_id(self):
        return self.__dict__['entity_id']

    @property
    def entity_version(self):
        return self.__dict__['entity_version']

    @property
    def timestamp(self):
        return self.__dict__['timestamp']

    def __eq__(self, rhs):
        if type(self) is not type(rhs):
            return NotImplemented
        return self.__dict__ == rhs.__dict__

    def __ne__(self, rhs):
        return not (self == rhs)

    def __hash__(self):
        return hash(tuple(itertools.chain(sorted(self.__dict__.items()), [type(self)])))

    def __repr__(self):
        return self.__class__.__qualname__ + "(" + ', '.join(
            "{0}={1!r}".format(*item) for item in sorted(self.__dict__.items())) + ')'


_event_handlers = {}


def subscribe(event_predicate, subscriber):
    if event_predicate not in _event_handlers:
        _event_handlers[event_predicate] = []
    _event_handlers[event_predicate].append(subscriber)


def unsubscribe(event_predicate, subscriber):
    if event_predicate in _event_handlers:
        handlers = _event_handlers[event_predicate]
        if subscriber in handlers:
            handlers.remove(subscriber)


def publish(event):
    matching_handlers = []
    for event_predicate, handlers in _event_handlers.items():
        if event_predicate(event):
            for handler in handlers:
                if handler not in matching_handlers:
                    matching_handlers.append(handler)
    for handler in matching_handlers:
        handler(event)



import unittest

try:
    from unittest import mock
except ImportError:
    import mock


from eventsourcing.domain.model.events import subscribe, publish, unsubscribe
from eventsourcing.domain.model.example import Example


class TestEvents(unittest.TestCase):

    def test_events(self):
        event = Example.Created(entity_id='entity1', a=1, b=2)

        # Check constructor keyword args lead to read-only attributes.
        self.assertEqual(1, event.a)
        self.assertEqual(2, event.b)
        self.assertRaises(AttributeError, getattr, event, 'c')
        self.assertRaises(AttributeError, setattr, event, 'c', 3)

        # Check domain event has auto-generated timestamp.
        self.assertIsInstance(event.timestamp, float)

        # Check timestamp value can be given to domain events.
        self.assertEqual(3, Example.Created(entity_id='entity1', a=1, b=2, timestamp=3).timestamp)

    def test_publish_subscribe(self):
        # Check subscribing event handlers with predicates.
        # - when predicate is True, handler should be called
        event = mock.Mock()
        predicate = mock.Mock()
        handler = mock.Mock()
        subscribe(event_predicate=predicate, subscriber=handler)
        publish(event)
        predicate.assert_called_once_with(event)
        handler.assert_called_once_with(event)

        # - when predicate is False, handler should NOT be called
        predicate = lambda x: False
        handler = mock.Mock()
        subscribe(event_predicate=predicate, subscriber=handler)
        publish(event)
        self.assertEqual(0, handler.call_count)

    def test_unsubscribe(self):
        # - when predicate is True, after unsubscribing, handler should NOT be called
        event = mock.Mock()
        predicate = mock.Mock()
        handler = mock.Mock()
        subscribe(event_predicate=predicate, subscriber=handler)
        unsubscribe(event_predicate=predicate, subscriber=handler)
        publish(event)
        predicate.assert_called_once_with(event)
        self.assertEqual(0, handler.call_count)

    def test_hash(self):
        event1 = Example.Created(entity_id='entity1', a=1, b=2, timestamp=3)
        event2 = Example.Created(entity_id='entity1', a=1, b=2, timestamp=3)
        self.assertEqual(hash(event1), hash(event2))

    def test_equality_comparison(self):
        event1 = Example.Created(entity_id='entity1', a=1, b=2, timestamp=3)
        event2 = Example.Created(entity_id='entity1', a=1, b=2, timestamp=3)
        event3 = Example.Created(entity_id='entity1', a=3, b=2, timestamp=3)
        self.assertEqual(event1, event2)
        self.assertNotEqual(event1, event3)
        self.assertNotEqual(event2, event3)
        self.assertNotEqual(event2, None)

    def test_repr(self):
        event1 = Example.Created(entity_id='entity1', a=1, b=2, timestamp=3)
        self.assertEqual("Example.Created(a=1, b=2, entity_id='entity1', entity_version=0, timestamp=3)", repr(event1))



import unittest

from eventsourcing.infrastructure.event_sourced_repos.example_repo import ExampleRepository
from eventsourcing.infrastructure.event_store import EventStore
from eventsourcing.infrastructure.stored_events import InMemoryStoredEventRepository
from eventsourcing.domain.model.example import Example


class TestEventSourcedRepository(unittest.TestCase):

    def test(self):
        # Setup an event store.
        stored_event_repo = InMemoryStoredEventRepository()
        event_store = EventStore(stored_event_repo=stored_event_repo)

        # Put an event in the event store.
        entity_id = 'entity1'
        event_store.append(Example.Created(entity_id=entity_id, a=1, b=2))

        # Setup an example repository.
        example_repo = ExampleRepository(event_store=event_store)

        # Check the repo has the example.
        self.assertIn(entity_id, example_repo)
        self.assertNotIn('xxxxxxxx', example_repo)

        # Check the entity attributes.
        example = example_repo[entity_id]
        self.assertEqual(1, example.a)
        self.assertEqual(2, example.b)
        self.assertEqual(entity_id, example.id)



import datetime

if hasattr(datetime, 'timezone'):
    utc_timezone = datetime.timezone.utc
else:
    from datetime import tzinfo, timedelta


    class UTC(tzinfo):
        def utcoffset(self, date_time):
            return timedelta(0)
        def dst(self, date_time):
            return timedelta(0)
    utc_timezone = UTC()




def utc_now():
    now_datetime = datetime.datetime.now(utc_timezone)
    try:
        now_timestamp = now_datetime.timestamp()
    except AttributeError:
        now_timestamp = (now_datetime - datetime.datetime(1970, 1, 1, tzinfo=utc_timezone)).total_seconds()
    return now_timestamp



from sqlalchemy import create_engine
from sqlalchemy.ext.declarative.api import declarative_base
from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy.orm.scoping import ScopedSession
from sqlalchemy.sql.schema import Column, Sequence
from sqlalchemy.sql.sqltypes import Integer, String
from eventsourcing.infrastructure.stored_events import StoredEventRepository, StoredEvent


def get_scoped_session_facade(uri=None):
    if uri is None:
        uri = 'sqlite:///:memory:'
    engine = create_engine(uri)
    Base.metadata.create_all(engine)
    session_factory = sessionmaker(bind=engine)
    scoped_session_facade = scoped_session(session_factory)
    return scoped_session_facade


Base = declarative_base()


class SqlStoredEvent(Base):

    __tablename__ = 'stored_events'

    id = Column(Integer, Sequence('stored_event_id_seq'), primary_key=True)
    event_id = Column(String(), index=True)
    entity_id = Column(String(), index=True)
    event_topic = Column(String(), index=True)
    event_attrs = Column(String())

    # def __repr__(self):
    #     return "<SqlStoredEvent(id='%s', event_id='%s', entity_id='%s', event_topic='%s', event_attrs='%s')>" % (
    #         self.id, self.event_id, self.entity_id, self.event_topic, self.event_attrs)


def stored_from_sql(sql_stored_event):
    assert isinstance(sql_stored_event, SqlStoredEvent)
    return StoredEvent(
        event_id=sql_stored_event.event_id,
        entity_id=sql_stored_event.entity_id,
        event_attrs=sql_stored_event.event_attrs,
        event_topic=sql_stored_event.event_topic
    )


def sql_from_stored(stored_event):
    assert isinstance(stored_event, StoredEvent)
    return SqlStoredEvent(
        event_id=stored_event.event_id,
        entity_id=stored_event.entity_id,
        event_attrs=stored_event.event_attrs,
        event_topic=stored_event.event_topic
    )


class SQLAlchemyStoredEventRepository(StoredEventRepository):

    def __init__(self, db_session):
        assert isinstance(db_session, ScopedSession)
        self.db_session = db_session
        self._by_id = {}
        self._by_entity_id = {}
        self._by_topic = {}

    def append(self, stored_event):
        sql_stored_event = sql_from_stored(stored_event)
        try:
            self.db_session.add(sql_stored_event)
            self.db_session.commit()
        except:
            self.db_session.rollback()
            raise
        finally:
            self.db_session.close() # Begins a new transaction


    def __contains__(self, item):
        return bool(self.db_session.query(SqlStoredEvent).filter_by(event_id=item).count())

    def __getitem__(self, item):
        sql_stored_event = self.db_session.query(SqlStoredEvent).filter_by(event_id=item).first()
        return stored_from_sql(sql_stored_event)

    def get_entity_events(self, entity_id):
        sql_stored_events = self.db_session.query(SqlStoredEvent).filter_by(entity_id=entity_id)
        return map(stored_from_sql, sql_stored_events)

    def get_topic_events(self, event_topic):
        sql_stored_events = self.db_session.query(SqlStoredEvent).filter_by(event_topic=event_topic)
        return map(stored_from_sql, sql_stored_events)



from eventsourcing.domain.model.events import DomainEvent
from eventsourcing.infrastructure.stored_events import StoredEventRepository, serialize_domain_event


class EventStore(object):

    def __init__(self, stored_event_repo):
        assert isinstance(stored_event_repo, StoredEventRepository)
        self.stored_event_repo = stored_event_repo

    def append(self, domain_event):
        assert isinstance(domain_event, DomainEvent)
        stored_event = serialize_domain_event(domain_event)
        self.stored_event_repo.append(stored_event)



from abc import ABCMeta, abstractmethod
from collections import namedtuple
import importlib
import json
import uuid
from six import with_metaclass

from eventsourcing.domain.model.events import DomainEvent
from eventsourcing.exceptions import TopicResolutionError

StoredEvent = namedtuple('StoredEvent', ['event_id', 'entity_id', 'event_topic', 'event_attrs'])


class StoredEventRepository(with_metaclass(ABCMeta)):

    @abstractmethod
    def append(self, stored_event):
        """Saves given stored event in this repository.
        """

    @abstractmethod
    def __getitem__(self, item):
        """Returns stored event for given event ID.
        """

    @abstractmethod
    def __contains__(self, item):
        """Tests whether given event ID exists.
        """

    @abstractmethod
    def get_entity_events(self, entity_id):
        """Returns all events for given entity ID.
        """

    @abstractmethod
    def get_topic_events(self, event_topic):
        """Returns all events for given topic.
        """


class InMemoryStoredEventRepository(StoredEventRepository):

    def __init__(self):
        self._by_id = {}
        self._by_entity_id = {}
        self._by_topic = {}

    def append(self, stored_event):
        assert isinstance(stored_event, StoredEvent)
        event_id = stored_event.event_id
        entity_id = stored_event.entity_id
        topic = stored_event.event_topic

        # Store the event by event ID.
        self._by_id[event_id] = stored_event

        # Append the event to list for entity ID.
        if entity_id not in self._by_entity_id:
            self._by_entity_id[entity_id] = []
        self._by_entity_id[entity_id].append(stored_event)

        # Append the event to list for event topic.
        if topic not in self._by_topic:
            self._by_topic[topic] = []
        self._by_topic[topic].append(stored_event)

    def __getitem__(self, item):
        return self._by_id[item]

    def __contains__(self, item):
        return item in self._by_id

    def get_entity_events(self, entity_id):
        if entity_id not in self._by_entity_id:
            return []
        else:
            return self._by_entity_id[entity_id]

    def get_topic_events(self, event_topic):
        return self._by_topic[event_topic]


def serialize_domain_event(domain_event):
    """Serializes a domain event into a stored event.

    Args:
        domain_event: A domain event object.

    Returns:
        A stored event object.

    Raises:
        AssertionError: If the domain event object is not an instance of DomainEvent.
    """
    assert isinstance(domain_event, DomainEvent)
    event_id = uuid.uuid4().hex
    entity_id = domain_event.entity_id
    event_topic = topic_from_domain_event(domain_event)
    event_attrs = json.dumps(domain_event.__dict__, separators=(',', ':'), sort_keys=True)
    return StoredEvent(
        event_id=event_id,
        entity_id=entity_id,
        event_topic=event_topic,
        event_attrs=event_attrs,
    )


def recreate_domain_event(stored_event):
    """Recreates original domain event from stored event topic and event attrs.

    Args:
        stored_event: A stored domain event object (instance of StoredEvent).

    Returns:
        A domain event object (instance of DomainEvent).

    Raises:
        AssertionError: If the stored event is not an instance of StoredEvent.
        TopicResolutionError: If the stored event topic could not be resolved.
        ValueError: If the stored event data could not be parsed as JSON.
    """
    assert isinstance(stored_event, StoredEvent)
    event_class = resolve_event_topic(stored_event.event_topic)
    event_data = json.loads(stored_event.event_attrs)
    try:
        domain_event = event_class(**event_data)
    except TypeError:
        raise TypeError("Unable to instantiate class '{}' with data '{}'".format(stored_event.event_topic, stored_event.event_attrs))
    return domain_event


def topic_from_domain_event(domain_event):
    """Returns a string describing a domain event class.

    Args:
        domain_event: A domain event object.

    Returns:
        A string describing the domain event object's class.
    """
    assert isinstance(domain_event, DomainEvent)
    return domain_event.__module__ + '#' + domain_event.__class__.__qualname__


def resolve_event_topic(topic):
    """Return domain event class described by given topic.

    Args:
        topic: A string describing a domain event class.

    Returns:
        A domain event class.

    Raises:
        TopicResolutionError: If there is no such domain event class.
    """
    # Todo: Fix up this block to show what the topic is, and where it went wrong.
    try:
        module_name, _, class_name = topic.partition('#')
        module = importlib.import_module(module_name)
    except ImportError:
        raise TopicResolutionError()
    try:
        cls = resolve_attr(module, class_name)
    except AttributeError:
        raise TopicResolutionError()
    return cls


def resolve_attr(obj, path):
    """A recursive version of getattr for navigating dotted paths.

    Args:
        obj: An object for which we want to retrieve a nested attribute.
        path: A dot separated string containing zero or more attribute names.

    Returns:
        The attribute referred to by obj.a1.a2.a3...

    Raises:
        AttributeError: If there is no such attribute.
    """
    if not path:
        return obj
    head, _, tail = path.partition('.')
    head_obj = getattr(obj, head)
    return resolve_attr(head_obj, tail)



import unittest
from eventsourcing.exceptions import TopicResolutionError
from eventsourcing.infrastructure.stored_events import serialize_domain_event, recreate_domain_event, \
    resolve_event_topic, StoredEvent, InMemoryStoredEventRepository
from eventsourcing.infrastructure.stored_events_sqlalchemy import SQLAlchemyStoredEventRepository, \
    get_scoped_session_facade
from eventsourcing.domain.model.example import Example


class TestStoredEvent(unittest.TestCase):

    def test_serialize_domain_event(self):
        event1 = Example.Created(a=1, b=2, entity_id='entity1', timestamp=3)
        stored_event = serialize_domain_event(event1)
        self.assertEqual('entity1', stored_event.entity_id)
        self.assertEqual('eventsourcing.domain.model.example#Example.Created', stored_event.event_topic)
        self.assertEqual('{"a":1,"b":2,"entity_id":"entity1","entity_version":0,"timestamp":3}', stored_event.event_attrs)

    def test_recreate_domain_event(self):
        stored_event = StoredEvent(event_id='1',
                                   entity_id='entity1',
                                   event_topic='eventsourcing.domain.model.example#Example.Created',
                                   event_attrs='{"a":1,"b":2,"entity_id":"entity1","timestamp":3}')
        domain_event = recreate_domain_event(stored_event)
        self.assertIsInstance(domain_event, Example.Created)
        self.assertEqual('entity1', domain_event.entity_id)
        self.assertEqual(1, domain_event.a)
        self.assertEqual(2, domain_event.b)
        self.assertEqual(3, domain_event.timestamp)

        # Check the TypeError is raised.
        stored_event = StoredEvent(event_id='1',
                                   entity_id='entity1',
                                   event_topic='os#path',
                                   event_attrs='{"a":1,"b":2,"entity_id":"entity1","timestamp":3}')
        self.assertRaises(TypeError, recreate_domain_event, stored_event)

    def test_resolve_event_topic(self):
        example_topic = 'eventsourcing.domain.model.example#Example.Created'
        actual = resolve_event_topic(example_topic)
        self.assertEqual(Example.Created, actual)
        example_topic = 'xxxxxxxxxxxxx#Example.Event'
        self.assertRaises(TopicResolutionError, resolve_event_topic, example_topic)
        example_topic = 'eventsourcing.domain.model.example#Xxxxxxxx.Xxxxxxxx'
        self.assertRaises(TopicResolutionError, resolve_event_topic, example_topic)


class StoredEventRepositoryTestCase(unittest.TestCase):

    def assertStoredEventRepositoryImplementation(self, stored_event_repo):
        # Store an event for 'entity1'.
        stored_event1 = StoredEvent(event_id='1',
                                    entity_id='entity1',
                                    event_topic='eventsourcing.domain.model.example#Example.Created',
                                    event_attrs='{"a":1,"b":2,"entity_id":"entity1","timestamp":3}')
        stored_event_repo.append(stored_event1)

        # Check the repo contains the event.
        self.assertIn(stored_event1.event_id, stored_event_repo)  # __contains__
        self.assertIsInstance(stored_event_repo[stored_event1.event_id], StoredEvent)  # __getitem__
        self.assertIsInstance(repr(stored_event_repo[stored_event1.event_id]), str)  # __getitem__

        # Store another event for 'entity1'.
        stored_event2 = StoredEvent(event_id='2',
                                    entity_id='entity1',
                                    event_topic='eventsourcing.domain.model.example#Example.Created',
                                    event_attrs='{"a":1,"b":2,"entity_id":"entity1","timestamp":4}')
        stored_event_repo.append(stored_event2)

        # Get all events for 'entity1'.
        events = stored_event_repo.get_entity_events('entity1')
        events = list(events)  # Make sequence from the iterator.
        self.assertEqual(2, len(list(events)))
        # - check the first event
        self.assertIsInstance(events[0], StoredEvent)
        self.assertEqual(stored_event1.event_topic, events[0].event_topic)
        self.assertEqual(stored_event1.event_attrs, events[0].event_attrs)
        # - check the second event
        self.assertIsInstance(events[1], StoredEvent)
        self.assertEqual(stored_event2.event_topic, events[1].event_topic)
        self.assertEqual(stored_event2.event_attrs, events[1].event_attrs)

        # Get all events for the topic.
        events = stored_event_repo.get_topic_events('eventsourcing.domain.model.example#Example.Created')
        events = list(events)
        self.assertEqual(2, len(events))
        # - check the first event
        self.assertIsInstance(events[0], StoredEvent)
        self.assertEqual(stored_event1.event_topic, events[0].event_topic)
        self.assertEqual(stored_event1.event_attrs, events[0].event_attrs)
        # - check the second event
        self.assertIsInstance(events[1], StoredEvent)
        self.assertEqual(stored_event2.event_topic, events[1].event_topic)
        self.assertEqual(stored_event2.event_attrs, events[1].event_attrs)


class TestInMemoryStoredEventRepository(StoredEventRepositoryTestCase):

    def test(self):
        self.assertStoredEventRepositoryImplementation(InMemoryStoredEventRepository())


class TestSQLAlchemyStoredEventRepository(StoredEventRepositoryTestCase):

    def test(self):
        stored_event_repo = SQLAlchemyStoredEventRepository(get_scoped_session_facade())
        self.assertStoredEventRepositoryImplementation(stored_event_repo)



from eventsourcing.application.main import EventSourcedApplication
from eventsourcing.domain.model.example import register_new_example
from eventsourcing.infrastructure.event_sourced_repos.example_repo import ExampleRepository


class ExampleApplication(EventSourcedApplication):
    """
    Example event sourced application.

    This application has an Example repository, and a factory method for
    registering new examples. It inherits an event store, a persistence
    subscriber, and a stored event repository, and a database connection.
    """
    def __init__(self):
        super(ExampleApplication, self).__init__()
        self.example_repo = ExampleRepository(event_store=self.event_store)

    def register_new_example(self, a, b):
        return register_new_example(a=a, b=b)



from distutils.core import setup

# Do we need to install mock?
install_requires_mock = False
try:
    from unittest import mock
except ImportError:
    try:
        import mock
    except ImportError:
        install_requires_mock = True


setup(
    name='eventsourcing',
    version='0.7.0',
    description='Event sourcing in Python',
    author='John Bywater',
    author_email='john.bywater@appropriatesoftware.net',
    url = 'https://github.com/johnbywater/eventsourcing',
    packages=[
        'eventsourcing',
        'eventsourcing/application',
        'eventsourcing/domain',
        'eventsourcing/domain/model',
        'eventsourcing/infrastructure',
        'eventsourcing/infrastructure/event_sourced_repos',
        'eventsourcing/utils',
        'eventsourcingtests',
    ],
    install_requires=['six'] + ['mock'] if install_requires_mock else [''],
    extras_require={
        'sqlalchemy': ['sqlalchemy'],
    },
    zip_safe=False,
    keywords=['eventsourcing', 'ddd', 'cqrs'],
    classifiers=[],
)



import unittest

from eventsourcing.infrastructure.event_player import EventPlayer
from eventsourcing.infrastructure.event_store import EventStore
from eventsourcing.infrastructure.stored_events import InMemoryStoredEventRepository
from eventsourcing.domain.model.example import Example


class TestEventPlayer(unittest.TestCase):

    def test(self):
        stored_event_repo = InMemoryStoredEventRepository()
        event_store = EventStore(stored_event_repo)
        # Store example events.
        event1 = Example.Created(entity_id='entity1', timestamp=3, a=1, b=2)
        event2 = Example.Created(entity_id='entity2', timestamp=4, a=2, b=4)
        event3 = Example.Created(entity_id='entity3', timestamp=5, a=3, b=6)
        event4 = Example.Discarded(entity_id='entity3', timestamp=6, entity_version=1)
        event_store.append(event1)
        event_store.append(event2)
        event_store.append(event3)
        event_store.append(event4)
        # Check the event sourced entities are correct.
        # - just use a trivial mutator that always instantiates the 'Example'.
        event_player = EventPlayer(event_store=event_store, domain_class=Example)
        self.assertEqual('entity1', event_player['entity1'].id)
        self.assertEqual(1, event_player['entity1'].a)
        self.assertEqual(2, event_player['entity2'].a)

        # Check entity3 is not found.
        self.assertRaises(KeyError, event_player.__getitem__, 'entity3')
